@@ -1,5 +1,5 @@
 package App::gimpgitbuild::Command::build;
-$App::gimpgitbuild::Command::build::VERSION = '0.18.1';
+$App::gimpgitbuild::Command::build::VERSION = '0.20.0';
 use strict;
 use warnings;
 use 5.014;
@@ -12,6 +12,18 @@ use Path::Tiny qw/ path tempdir tempfile cwd /;
 use App::gimpgitbuild::API::GitBuild ();
 use Git::Sync::App                   ();
 
+sub _api_obj
+{
+    my $self = shift;
+
+    if (@_)
+    {
+        $self->{_api_obj} = shift;
+    }
+
+    return $self->{_api_obj};
+}
+
 sub description
 {
     return "build gimp from git";
@@ -57,8 +69,8 @@ BEGIN
         $skip_builds_re = qr/$re_str/;
     }
 }
-my $MESON_BUILD_DIR = ( $ENV{GIMPGITBUILD__MESON_BUILD_DIR}
-        // "to-del--gimpgitbuild--meson-build" );
+my $BUILD_DIR = ( $ENV{GIMPGITBUILD__MESON_BUILD_DIR}
+        // "to-del--gimpgitbuild--build-dir" );
 
 # See:
 # https://github.com/libfuse/libfuse/issues/212
@@ -88,32 +100,56 @@ sub _git_build
     $args->{branch} //= 'master';
     $args->{tag}    //= 'false';
 
-    my $git_co = $args->{git_co};
-    if ( !-e "$args->{git_co}" )
+    my $git_co = (
+        $args->{git_co} // (
+                  $self->_api_obj()->base_git_clones_dir() . "/"
+                . $args->{git_checkout_subdir}
+        )
+    );
+    if ( !-e $git_co )
     {
-        path( $args->{git_co} )->parent->mkpath;
+        path($git_co)->parent->mkpath;
         _do_system( { cmd => [qq#git clone "$args->{url}" "$git_co"#] } );
     }
 
     my $meson_build_shell_cmd =
-qq#mkdir -p "$MESON_BUILD_DIR" && cd "$MESON_BUILD_DIR" && meson --prefix="$args->{prefix}" $UBUNTU_MESON_LIBDIR_OVERRIDE .. && ninja $PAR_JOBS && ninja $PAR_JOBS test && ninja $PAR_JOBS install#;
+qq#mkdir -p "$BUILD_DIR" && cd "$BUILD_DIR" && meson --prefix="$args->{prefix}" $UBUNTU_MESON_LIBDIR_OVERRIDE .. && ninja $PAR_JOBS && ninja $PAR_JOBS test && ninja $PAR_JOBS install#;
     my $autoconf_build_shell_cmd =
-qq#NOCONFIGURE=1 ./autogen.sh && ./configure @{$extra_configure_args} --prefix="$args->{prefix}" && make $PAR_JOBS && @{[_check()]} && make install#;
+qq#NOCONFIGURE=1 ./autogen.sh && mkdir -p "$BUILD_DIR" && cd "$BUILD_DIR" && ../configure @{$extra_configure_args} --prefix="$args->{prefix}" && make $PAR_JOBS && @{[_check()]} && make install#;
     my $sync_cmd = $self->_git_sync( { branch => $args->{branch}, } );
-    _do_system(
-        {
-            cmd => [
+    my $run      = sub {
+        return _do_system(
+            {
+                cmd => [
 qq#cd "$git_co" && git checkout "$args->{branch}" && ( $args->{tag} || $sync_cmd ) && #
-                    . (
-                    ( $self->{mode} eq 'clean' ) ? "git clean -dxf ."
-                    : (
-                          $args->{use_meson} ? $meson_build_shell_cmd
-                        : $autoconf_build_shell_cmd
-                    )
-                    )
-            ]
+                        . (
+                        ( $self->{mode} eq 'clean' ) ? "git clean -dxf ."
+                        : (
+                              $args->{use_meson} ? $meson_build_shell_cmd
+                            : $autoconf_build_shell_cmd
+                        )
+                        )
+                ]
+            }
+        );
+    };
+
+    my $on_failure = $args->{on_failure};
+
+    if ( !$on_failure )
+    {
+        $run->();
+    }
+    else
+    {
+        eval { $run->(); };
+        my $Err = $@;
+
+        if ($Err)
+        {
+            $on_failure->( { exception => $Err, }, );
         }
-    );
+    }
     return;
 }
 
@@ -155,6 +191,7 @@ sub execute
 
     my $fh  = \*STDIN;
     my $obj = App::gimpgitbuild::API::GitBuild->new;
+    $self->_api_obj($obj);
 
     my $HOME = $obj->home_dir;
     my $env  = $obj->new_env;
@@ -164,42 +201,41 @@ sub execute
     _which_xvfb_run();
     _ascertain_lack_of_gtk_warnings();
     $self->{mode} = $mode;
-    my $base_src_dir = $obj->base_git_clones_dir;
 
     my $GNOME_GIT = 'https://gitlab.gnome.org/GNOME';
     $self->_git_build(
         {
-            id        => "babl",
-            git_co    => "$base_src_dir/babl/git/babl",
-            url       => "$GNOME_GIT/babl",
-            prefix    => $obj->babl_p,
-            use_meson => 1,
+            id                  => "babl",
+            git_checkout_subdir => "babl/git/babl",
+            url                 => "$GNOME_GIT/babl",
+            prefix              => $obj->babl_p,
+            use_meson           => 1,
         }
     );
     $self->_git_build(
         {
-            id        => "gegl",
-            git_co    => "$base_src_dir/gegl/git/gegl",
-            url       => "$GNOME_GIT/gegl",
-            prefix    => $obj->gegl_p,
-            use_meson => 1,
+            id                  => "gegl",
+            git_checkout_subdir => "gegl/git/gegl",
+            url                 => "$GNOME_GIT/gegl",
+            prefix              => $obj->gegl_p,
+            use_meson           => 1,
         }
     );
     $self->_git_build(
         {
-            id        => "libmypaint",
-            git_co    => "$base_src_dir/libmypaint/git/libmypaint",
-            url       => "https://github.com/mypaint/libmypaint.git",
-            prefix    => $obj->mypaint_p,
-            use_meson => 0,
-            branch    => "v1.6.1",
-            tag       => "true",
+            id                  => "libmypaint",
+            git_checkout_subdir => "libmypaint/git/libmypaint",
+            url                 => "https://github.com/mypaint/libmypaint.git",
+            prefix              => $obj->mypaint_p,
+            use_meson           => 0,
+            branch              => "v1.6.1",
+            tag                 => "true",
         }
     );
     $self->_git_build(
         {
-            id        => "mypaint-brushes",
-            git_co    => "$base_src_dir/libmypaint/git/mypaint-brushes",
+            id                  => "mypaint-brushes",
+            git_checkout_subdir => "libmypaint/git/mypaint-brushes",
             url       => "https://github.com/Jehan/mypaint-brushes.git",
             prefix    => $obj->mypaint_p,
             use_meson => 0,
@@ -207,18 +243,34 @@ sub execute
         }
     );
 
-    my $KEY        = 'GIMPGITBUILD__BUILD_GIMP_USING_MESON';
-    my $GIMP_BUILD = ( exists( $ENV{$KEY} ) ? $ENV{$KEY} : 1 );
+    my $KEY                    = 'GIMPGITBUILD__BUILD_GIMP_USING_MESON';
+    my $BUILD_GIMP_USING_MESON = ( exists( $ENV{$KEY} ) ? $ENV{$KEY} : 1 );
 
-# autoconf_git_build "$base_src_dir/git/gimp" "$GNOME_GIT"/gimp "$HOME/apps/gimp-devel"
     $self->_git_build(
         {
             id                   => "gimp",
             extra_configure_args => [ qw# --enable-debug #, ],
-            git_co               => "$base_src_dir/git/gimp",
+            git_checkout_subdir  => "git/gimp",
             url                  => "$GNOME_GIT/gimp",
             prefix               => $obj->gimp_p,
-            use_meson            => $GIMP_BUILD,
+            use_meson            => $BUILD_GIMP_USING_MESON,
+            on_failure           => sub {
+                my ($args) = @_;
+                my $Err = $args->{exception};
+                if ( !$BUILD_GIMP_USING_MESON )
+                {
+                    die $Err;
+                }
+                STDERR->print( $Err, "\n" );
+                STDERR->print(<<"EOF");
+Meson-using builds of GIMP are known to be error prone. Please try setting
+the "$KEY" environment variable to "0", and run gimpgitbuild again, e.g using:
+
+    export $KEY="0"
+
+EOF
+                die "Meson build failure";
+            },
         }
     );
 
@@ -238,7 +290,7 @@ __END__
 
 =head1 VERSION
 
-version 0.18.1
+version 0.20.0
 
 =begin foo return (
         [ "output|o=s", "Output path" ],
